defmodule HXL.Lexer.Template do
  import NimbleParsec

  @newline ?\n
  @carriage_return ?\r
  @space ?\s
  @tab ?\t

  digit = ascii_char([?0..?9])
  non_zero_digit = ascii_char([?1..?9])
  negative_sign = ascii_char([?-])

  whitespace = ascii_char([@space, @tab])

  line_end =
    choice([
      ascii_char([@newline]),
      ascii_char([@carriage_return]) |> optional(ascii_char([@newline]))
    ])

  blankspace = choice([whitespace, line_end])

  ignoreed = ignore(blankspace)

  operators_keywords =
    choice([
      choice([
        string("&&"),
        string("||"),
        string("=="),
        string("!="),
        string("<="),
        string(">="),
        string("=>")
      ]),
      # Keywords
      # Asserts so that there isn't any identifiers after the keyword
      choice([
        string("null"),
        string("true"),
        string("false"),
        string("for"),
        string("if"),
        string("endif"),
        string("else"),
        string("in")
      ])
      |> lookahead_not(ascii_char([?a..?z, ?A..?Z, ?0..?9])),
      ascii_char([
        ?=,
        ?!,
        ??,
        ?+,
        ?-,
        ?*,
        ?/,
        ?%,
        ?$,
        ?<,
        ?>,
        ?.,
        ?,,
        ?:,
        ?*,
        ?{,
        ?},
        ?[,
        ?],
        ?(,
        ?)
      ]),
      times(ascii_char([?.]), 3)
    ])
    |> post_traverse({:atom_token, []})

  template_literal =
    repeat(
      lookahead_not(
        choice([
          blankspace,
          operators_keywords
        ])
      )
      |> utf8_char([])
    )
    |> reduce({List, :to_string, []})
    |> post_traverse({:labeled_token, [:literal]})

  template =
    choice([
      ignoreed,
      operators_keywords,
      template_literal
    ])

  defcombinator(:template, template)
  defparsec(:tokenize, repeat(lookahead_not(eos()) |> parsec(:template)))

  defp atom_token(_rest, [chars], context, loc, byte_offset) when is_binary(chars) do
    token_atom = String.to_atom(chars)
    {[{token_atom, line_and_column(loc, byte_offset, String.length(chars))}], context}
  end

  defp atom_token(_rest, chars, context, loc, byte_offset) do
    value = chars |> Enum.reverse()
    token_atom = value |> List.to_atom()
    {[{token_atom, line_and_column(loc, byte_offset, length(value))}], context}
  end

  defp labeled_token(_rest, chars, context, loc, byte_offset, token_name) do
    value = chars |> Enum.reverse()

    {[{token_name, line_and_column(loc, byte_offset, length(value)), value}], context}
  end

  def line_and_column({line, line_offset}, byte_offset, column_correction) do
    column = byte_offset - line_offset - column_correction + 1
    {line, column}
  end

  defp do_label_token(token_name, value, loc, byte_offset) do
    {token_name, line_and_column(loc, byte_offset, length(value)), value}
  end
end
